//----------------------------------------------------------------------------------------
// map_info.H: Contains declatrations for classes defining mapping from reference to
//             physical elements, edges, faces
//----------------------------------------------------------------------------------------

#ifndef MAP_INFO_H
#define MAP_INFO_H

#include <cmath>
#include <fstream>
#include "general.H"

//--------------------------------------------------------------------------------------
// MAP Class: Abstract base class: Defines common interface for all MAP classes
//--------------------------------------------------------------------------------------

class MAP
{
protected:
  unsigned char d_space_in, d_space_out; // Dimensions of reference and physical spaces

public:
  // Default constructor
  // Parameters: in - dimension of reference space; out - dimension of physical space
  MAP(int in, int out) {d_space_in=in; d_space_out=out;}

  // Map from reference to physical space
  // Parameters: in - coordinates in reference space; out - coordinates in physical space
  virtual void mapto(const real in[], real out[]) const = 0;

  // Map from physical to reference space
  // Parameters: in - coordinates in physical space; out - coordinates in reference space
  virtual void mapto_inverse(const real in[], real out[]) const = 0;
};



//--------------------------------------------------------------------------------------
// MapEdge Class: Map from [0, 1] to an edge in 2D
//--------------------------------------------------------------------------------------

class MapEdge : public MAP
{
protected:
  int ind;                                     // index of dimension used for inverse map
  real ref_point[d_space_2d], map[d_space_2d]; // out[i]=map[i]*in[0]+ref_point[i]
  long_real length;                            // edge length

public:
  // Default constructor
  MapEdge() : MAP(1, 2) {for (int i=0; i<d_space_2d; ref_point[i]=map[i]=0., i++); length=0.;}

  // Compute affine mapping coefficients from reference to physical space
  // Parameters: coord - edge node coordinates
  void UpdateMap(const real* coord[d_space]);

  // Map from reference to physical space
  // Parameters: in - coordinates in reference space; out - coordinates in physical space
  void mapto(const real in[], real out[]) const 
  {out[0]=map[0]*in[0]+ref_point[0]; out[1]=map[1]*in[0]+ref_point[1];}

  // Map from physical to reference space
  // Parameters: in - coordinates in physical space; out - coordinates in reference space
  void mapto_inverse(const real in[], real out[]) const 
  {out[0]=(in[ind]-ref_point[ind])/map[ind];}

  // Return value: length of the edge
  long_real GetScale() const {return length;}
};



//--------------------------------------------------------------------------------------
// MapTriangle Class: Map from {(0, 0), (0, 1), (1, 0)} to a triangle in 2D
//--------------------------------------------------------------------------------------

class MapTriangle : public MAP // Map from {(0, 0), (0, 1), (1, 0)} to a triangle in 2D
{
protected:
  real ref_point[d_space_2d], map[d_space_2d][d_space_2d]; // out[i]=map[i][j]*in[j]+ref_point[i]
  real ref_point_inv[d_space_2d], map_inv[d_space_2d][d_space_2d];
  long_real scale, diam; // scale - absolute value of Jacobian determinant; diam - diameter of the triangle

public:
  // Default constructor
  MapTriangle() : MAP(2, 2)
  {for (int i=0; i<d_space_2d; ref_point[i]=ref_point_inv[i]=0., i++)
  for (int j=0; j<d_space_2d; map[i][j]=map_inv[i][j]=0., j++); scale=diam=0.;}

  // Compute affine mapping coefficients from reference to physical space
  // Parameters: coord - triangle node coordinates
  void UpdateMap(const real* coord[d_space]);

  // Map from reference to physical space
  // Parameters: in - coordinates in reference space; out - coordinates in physical space
  void mapto(const real in[], real out[]) const 
  {out[0]=map[0][0]*in[0]+map[0][1]*in[1]+ref_point[0]; 
  out[1]=map[1][0]*in[0]+map[1][1]*in[1]+ref_point[1];}

  // Map from physical to reference space
  // Parameters: in - coordinates in physical space; out - coordinates in reference space
  void mapto_inverse(const real in[], real out[]) const 
  {out[0]=map_inv[0][0]*in[0]+map_inv[0][1]*in[1]+ref_point_inv[0]; 
  out[1]=map_inv[1][0]*in[0]+map_inv[1][1]*in[1]+ref_point_inv[1];}

  // Return value: absolute value of Jacobian determinant
  long_real GetScale() const {return scale;}

  // Return value: diameter of the element
  long_real GetDiameter() const {return diam;}

  // Map the gradient from reference to physical space
  // Parameters: in - gradient in reference space; out - gradient in physical space
  void MapGradient(const real in[], real out[]) const
  {out[0]=map_inv[0][0]*in[0]+map_inv[1][0]*in[1];
  out[1]=map_inv[0][1]*in[0]+map_inv[1][1]*in[1];}
};



//--------------------------------------------------------------------------------------
// MapHorizFace Class: Map from {(0, 0), (0, 1), (1, 0)} to a horizontal face of a prism
//--------------------------------------------------------------------------------------

class MapHorizFace : public MAP
{
protected:
  real ref_point, map[d_space_2d]; // out[2]=map[0]*in[0]+map[1]*in[1]+ref_point
  long_real scale;                 // scale - absolute value of Jacobian determinant
  MapTriangle map_2d;              // map for corresponding 2D triangle

public:
  // Default constructor
  MapHorizFace() : MAP(2, 3) {for (int i=0; i<d_space_2d; map[i]=0., i++); ref_point=scale=0.;}

  // Constructor
  // Parameters: map_2d_ - map for corresponding 2D triangle
  MapHorizFace(const MapTriangle *map_2d_) : MAP(2, 3)
  {map_2d=*map_2d_; for (int i=0; i<d_space_2d; map[i]=0., i++); ref_point=scale=0.;}

  // Compute affine mapping coefficients from reference to physical space
  // Parameters: coord - face node coordinates
  void UpdateMap(const real* coord[d_space]);

  // Map from reference to physical space
  // Parameters: in - coordinates in reference space; out - coordinates in physical space
  void mapto(const real in[], real out[]) const
  {map_2d.mapto(in, out); out[2]=map[0]*in[0]+map[1]*in[1]+ref_point;}

  // Map from physical to reference space
  // Parameters: in - coordinates in physical space; out - coordinates in reference space
  void mapto_inverse(const real in[], real out[]) const {map_2d.mapto_inverse(in, out);}

  // Return value: absolute value of Jacobian determinant
  long_real GetScale() const {return scale;}

  // Setting corresponding 2D triangle
  // Parameters: new_map - map for corresponding 2D triangle
  void Set2DMap(MapTriangle* new_map) {map_2d=*new_map;}
};



//--------------------------------------------------------------------------------------
// MapVertFace Class: Map from [0,1]x[0,1] to a vertical face of a prism
//--------------------------------------------------------------------------------------

class MapVertFace : public MAP
{
protected:
  // out[2]=map[0]*in[0]+map[1]*in[1]+bilin_coef*in[0]*in[1]+ref_point
  real ref_point, map[d_space_2d], bilin_coef;
  MapEdge map_2d;                             // map for corresponding 2D edge

public:
  // Default constructor
  MapVertFace() : MAP(2, 3)
  {for (int i=0; i<d_space_2d; map[i]=0., i++); ref_point=bilin_coef=0.;}

  // Constructor
  // Parameters: map_2d_ - map for corresponding 2D edge
  MapVertFace(const MapEdge *map_2d_) : MAP(2, 3)
  {map_2d=*map_2d_; for (int i=0; i<d_space_2d; map[i]=0., i++); ref_point=bilin_coef=0.;}

  // Compute affine mapping coefficients from reference to physical space
  // Parameters: coord - face node coordinates
  void UpdateMap(const real* coord[d_space]);

  // Map from reference to physical space
  // Parameters: in - coordinates in reference space; out - coordinates in physical space
  void mapto(const real in[], real out[]) const
  {map_2d.mapto(in, out); out[2]=map[0]*in[0]+(map[1]+bilin_coef*in[0])*in[1]+ref_point;}

  // Map from physical to reference space
  // Parameters: in - coordinates in physical space; out - coordinates in reference space
  void mapto_inverse(const real in[], real out[]) const {map_2d.mapto_inverse(in, out);
  out[1]=(in[2]-ref_point-map[0]*out[0])/(map[1]+bilin_coef*out[0]);}

  // Return value: area of the face
  long_real GetScale() const {return std::abs(map[1]+0.5*bilin_coef)*map_2d.GetScale();}

  // Return value: height*length at a point
  // Parameters: xi - coordinates in reference space
  long_real GetCorrection(const real xi[]) const 
  {return std::abs(map[1]+bilin_coef*xi[0])*map_2d.GetScale();}

  // Return value: height at a point
  // Parameters: xi - coordinates in reference space
  real GetHeight(const real xi[]) const 
  {return std::abs(map[1]+bilin_coef*xi[0]);}

  // Setting corresponding 2D edge
  // Parameters: new_map - map for corresponding 2D edge
  void Set2DMap(MapEdge* new_map) {map_2d=*new_map;}
};



//--------------------------------------------------------------------------------------
// MapPrism Class: Map from the reference to a physical prism
//--------------------------------------------------------------------------------------

class MapPrism : public MAP
{
protected:
  // out[2]=map[0]*in[0]+map[1]*in[1]+map[2]*in[2]
  //        +bilin_coef[0]*in[0]*in[2]+bilin_coef[1]*in[1]*in[2]+ref_point
  real ref_point, map[d_space], bilin_coef[d_space_2d];
  long_real vert_scale;       // Average height of the prism
  MapTriangle map_2d;         // map for corresponding 2D triangle

public:
  // Default constructor
  MapPrism() : MAP(3, 3) {for (int i=0; i<d_space; map[i]=0., i++);
  ref_point=bilin_coef[0]=bilin_coef[1]=vert_scale=0.;}

  // Constructor
  // Parameters: map_2d_ - map for corresponding 2D triangle
  MapPrism(const MapTriangle *map_2d_) : MAP(3, 3)
  {
    map_2d=*map_2d_;
    for (int i=0; i<d_space; map[i]=0., i++);
    ref_point=bilin_coef[0]=bilin_coef[1]=vert_scale=0.;
  }

  // Compute affine mapping coefficients from reference to physical space
  // Parameters: coord - prism node coordinates
  void UpdateMap(const real* coord[d_space]);

  // Map vertical coordinate from reference to physical space
  // Parameters: in - coordinates in reference space; out - coordinates in physical space
  long_real mapto_z(const real in[]) const {return map[0]*in[0]+map[1]*in[1]+ref_point+
      (bilin_coef[0]*in[0]+bilin_coef[1]*in[1]+map[2])*in[2];}

  // Map vertical coordinate from physical to reference space
  // Parameters: in - coordinates in physical space; out - coordinates in reference space
  long_real mapto_inverse_z(const real in[]) const
  {return (in[2]-ref_point-map[0]*in[0]-map[1]*in[1])/
      (map[2]+bilin_coef[0]*in[0]+bilin_coef[1]*in[1]);}

  // Map from reference to physical space
  // Parameters: in - coordinates in reference space; out - coordinates in physical space
  void mapto(const real in[], real out[]) const
  {map_2d.mapto(in, out); out[2]=mapto_z(in);}

  // Map from physical to reference space
  // Parameters: in - coordinates in physical space; out - coordinates in reference space
  void mapto_inverse(const real in[], real out[]) const
  {map_2d.mapto_inverse(in, out); out[2]=in[2]; out[2]=mapto_inverse_z(out);}

  // Return value: absolute value of Jacobian determinant
  long_real GetScale() const {return vert_scale*map_2d.GetScale();}

  // Return value: average height
  long_real GetVerticalScale() const {return vert_scale;}

  // Return value: height at the point
  // Parameters: xi - coordinates in reference space
  real GetHeight(const real xi[]) const
  {return std::abs(map[2]+bilin_coef[0]*xi[0]+bilin_coef[1]*xi[1]);}

  // Return value: ratio of height at the point to average height
  // Parameters: xi - coordinates in reference space
  long_real GetCorrection(const real xi[]) const {return GetHeight(xi)/vert_scale;}

  // Get scaling parameters used for physical basis objects
  // Parameters: scale - scaling; shift - vector to origin
  void GetPhysicalBasisParam(real scale[], real shift[]) const;

  // Map the gradient from reference to physical space
  // Parameters: in - gradient in reference space; out - gradient in physical space
  void MapGradient(const real in[], real out[]) const
  {map_2d.MapGradient(in, out); out[2]=in[2]/map[2];}

  // Setting corresponding 2D triangle
  // Parameters: new_map - map for corresponding 2D triangle
  const MapTriangle& Map2D() {return map_2d;}

  // Setting corresponding 2D triangle
  // Parameters: new_map - map for corresponding 2D triangle
  void Set2DMap(MapTriangle* new_map) {map_2d=*new_map;}

  // Return value: diameter of the element
  long_real GetDiameter() const {return std::sqrt(map_2d.GetDiameter()*vert_scale);}
};

#endif
