//----------------------------------------------------------------------------------------
// mesh_2d.H: Contains declarations for nodes, edges, and triangles in 2D
//----------------------------------------------------------------------------------------

#ifndef MESH_2D_H
#define MESH_2D_H

#include <fstream>
#include <cassert>
#include "general.H"
#include "map_info.H"

//--------------------------------------------------------------------------------------
// NODE_2D Class: Node class in 2D
//--------------------------------------------------------------------------------------

class NODE_2D
{
private:
  int id;                   // Unique id of the node
  real scale;               // Combined area of elements surrounding the node
  real coord[d_space_2d];   // Coordinates of the node

public:
  // Default constructor
  NODE_2D();

  // Constructor for hot start input
  // Parameters: hot start input file stream
  NODE_2D(std::ifstream& in);

  // Default destructor
  ~NODE_2D() {}

  // Return value: Node ID, rvalue
  int ID() const {return id;}

  // Return value: Node ID, lvalue
  int& ID() {return id;}

  // Return value: node coordinates
  const real* Coord() const {return coord;}

  // Parameters: i - coordinate index 
  // Return value: i-th node coordinate, rvalue
  real Coord(int i) const
  {
    assert_debug(i>=0 && i<d_space_2d);
    return coord[i];
  }

  // Parameters: i - coordinate index 
  // Return value: i-th node coordinate, lvalue
  real& Coord(int i)
  {
    assert_debug(i>=0 && i<d_space_2d);
    return coord[i];
  }

  // Combined area of elements surrounding the node, rvalue
  real Scale() const {return scale;}

  // Combined area of elements surrounding the node, lvalue
  real& Scale() {return scale;}

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream& out, const NODE_2D& node);
};



//--------------------------------------------------------------------------------------
// EDGE_2D Class: Edge class in 2D
//--------------------------------------------------------------------------------------

class EDGE_2D
{
private:
  // System data
  int id;                   // Unique id of the edge
  int type;                 // Edge type field
  int status;               // Status field

  // Geometry data
  MapEdge mapping;          // Map from the reference edge
  long_real normal[d_space_2d];  // Normal vector 

  // Connectivity data
  int node[2];              // Pointers to edge endpoints
  int element[2];           // Pointers to elements sharing the edge

  // Hierarchy data
  int parent;               // id of the parent edge
  int n_child;              // number of child edges
  int child;                // id of the first child edge

public:
  enum status_value {SPLIT=1, ACTIVE=2};

  // Default constructor
  EDGE_2D();

  // Constructor for hot start input
  // Parameters: hot start input file stream
  EDGE_2D(std::ifstream& in);

  // Default destructor
  ~EDGE_2D() {}

  // Compute mapping from the reference edge
  // Parameters: node_coord - node coordinates
  void ComputeReferenceMapping(const real *node_coord[2]);

  // Return value: Mapping from the reference edge
  const MapEdge& Mapping() const {return mapping;}

  // Return value: Edge ID, rvalue
  int ID() const {return id;}

  // Return value: Edge ID, lvalue
  int& ID() {return id;}

  // Return value: Edge type, rvalue
  int Type() const {return type;}

  // Return value: Edge type, lvalue
  int& Type() {return type;}

  // Return value: Status, rvalue
  int Status() const {return status;}

  // Return value: Status, lvalue
  int& Status() {return status;}

  // Return value: Normal vector
  const long_real* Normal() const {return normal;}

  // Parameters: i - coordinate index
  // Return value: i-th normal component, rvalue
  long_real Normal(int i) const
  {
    assert_debug(i>=0 && i<d_space_2d);
    return normal[i];
  }

  // Parameters: i - coordinate index
  // Return value: i-th normal component, lvalue
  long_real& Normal(int i)
  {
    assert_debug(i>=0 && i<d_space_2d);
    return normal[i];
  }

  // Parameters: i - node index
  // Return value: i-th node id, rvalue
  int Node(int i) const
  {
    assert_debug(i>=0 && i<2);
    return node[i];
  }

  // Parameters: i - node index
  // Return value: i-th node id, lvalue
  int& Node(int i)
  {
    assert_debug(i>=0 && i<2);
    return node[i];
  }

  // Parameters: i - element index
  // Return value: i-th element id, rvalue
  int Element(int i) const
  {
    assert_debug(i>=0 && i<2);
    return element[i];
  }

  // Parameters: i - element index
  // Return value: i-th element id, lvalue
  int& Element(int i)
  {
    assert_debug(i>=0 && i<2);
    return element[i];
  }

  // Parameters: i - element index
  // Return value: is i-th element present?
  bool IsElement(int i) const
  {
    assert_debug(i>=0 && i<2);
    return element[i]!=BAD_INDEX;
  }

  // Parameters: el_id - element id
  // Return value: element index if element is contiguous, BAD_INDEX otherwise
  int ElementIndex(int el_id) const
  {int ind = el_id==element[0] ? 0 : (el_id==element[1] ? 1 : BAD_INDEX); return ind;}

  // Return value: id of the parent edge (if present), rvalue
  int Parent() const {return parent;}

  // Return value: id of the parent edge (if present), lvalue
  int& Parent() {return parent;}

  // Parameters: n - number of children
  // Return value: id of the first child edge
  void SetChildren(int n, int first)
  {
    assert_debug(n>=0);
    n_child=n;
    child=first;
  }

  // Return value: number of children
  int NChildren() const  {return n_child;}

  // Parameters: i - index of the child edge
  // Return value: id of the i-th child edge
  int Child(int i) const
  {
    assert_debug(i>=0 && i<n_child);
    return child+i;
  }

  // Return value: is an active edge?
  bool IsActive() const {return (status&ACTIVE)!=0;}

  // Set the edge to be active
  void SetActive() {status|=ACTIVE;}

  // Set recursively the edge (or its descendants) to be active
  void SetActiveRecursive();

  // Set the edge to be inactive
  void UnSetActive() {status&=~ACTIVE;}

  // Set recursively the edge (or its descendants) to be inactive
  void UnSetActiveRecursive();

  // Return value: is the edge refined?
  bool IsRefined() const {return (status&SPLIT)!=0;}

  // Refine the edge using edge bisection
  // Parameters: edge_child - child edges; node - nodes on the edge; node_new - new node
  // Return value: id of the first child edge
  void Refine(EDGE_2D* edge_child[2], NODE_2D* node[], NODE_2D* node_new);

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream& out, const EDGE_2D& edge);
};



//--------------------------------------------------------------------------------------
// TRIANGLE_2D Class: Element class in 2D
//--------------------------------------------------------------------------------------

class TRIANGLE_2D
{
private:
  // System data
  int id;                          // Unique id of the element
  int type;                        // Element type field
  int status;                      // Status field

  // Geometry data
  MapTriangle mapping;             // Map from the reference element

  // Connectivity data
  int node[vert_h];                // Pointers to the element nodes
  int edge[vert_h];                // Pointers to the element edges
  char index[vert_h];              // Element index in the contiguous edges

  // Hierarchy data
  int parent;                      // id of the parent element
  int n_child;                     // Number of child elements
  int child;                       // id of the first child element
  unsigned char refinement_level;  // Level of refinement the element belongs to

public:
  enum status_value {SPLIT=1, ACTIVE=2, REFINE=4, UNREFINE=8};

  // Default constructor
  TRIANGLE_2D();

  // Constructor for hot start input
  // Parameters: hot start input file stream
  TRIANGLE_2D(std::ifstream& in);

  // Default destructor
  ~TRIANGLE_2D() {}

  // Compute mapping from the reference triangle
  // Parameters: node_coord - node coordinates
  void ComputeReferenceMapping(const real* node_coord[vert_h])
  {mapping.UpdateMap(node_coord);}

  // Return value: Mapping from the reference triangle
  const MapTriangle& Mapping() const {return mapping;}

  // Return value: Element ID, rvalue
  int ID() const {return id;}

  // Return value: Element ID, lvalue
  int& ID() {return id;}

  // Return value: Element type, rvalue
  int Type() const {return type;}

  // Return value: Element type, lvalue
  int& Type() {return type;}

  // Return value: Status, rvalue
  int Status() const {return status;}

  // Return value: Status, lvalue
  int& Status() {return status;}

  // Parameters: i - node index
  // Return value: i-th node id, rvalue
  int Node(int i) const
  {
    assert_debug(i>=0 && i<vert_h);
    return node[i];
  }

  // Parameters: i - node index
  // Return value: i-th node id, lvalue
  int& Node(int i)
  {
    assert_debug(i>=0 && i<vert_h);
    return node[i];
  }

  // Parameters: i - edge index
  // Return value: i-th edge id, rvalue
  int Edge(int i) const
  {
    assert_debug(i>=0 && i<vert_h);
    return edge[i];
  }

  // Parameters: i - edge index
  // Return value: i-th edge id, lvalue
  int& Edge(int i)
  {
    assert_debug(i>=0 && i<vert_h);
    return edge[i];
  }

  // Parameters: ed - edge id
  // Return value: edge index in the element
  int EdgeIndex(int ed) const
  {for (int i=0; i<vert_h; i++) if (ed==edge[i]) return i; return BAD_INDEX;}

  // Parameters: i - element index
  // Return value: element index in ith edge, rvalue
  char ElementIndex(int i) const
  {
    assert_debug(i>=0 && i<vert_h);
    return index[i];
  }

  // Parameters: i - element index
  // Return value: element index in ith edge, lvalue
  char& ElementIndex(int i)
  {
    assert_debug(i>=0 && i<vert_h);
    return index[i];
  }

  // Parameters: coord - point coordinates
  // Return value: Distance from coord[] to the barycenter of the element
  long_real DistanceToBarycenter(const real coord[]) const;

  // Parameters: coord - point coordinates
  // Return value: Distance from coord[] to the boundary of the element (0 if the point is inside)
  long_real DistanceToBoundary(const real coord[]) const;

  // Return value: id of the parent element (if present), rvalue
  int Parent() const {return parent;}

  // Return value: id of the parent element (if present), lvalue
  int& Parent() {return parent;}

  // Parameters: n - number of children
  // Return value: id of the first child element
  void SetChildren(int n, int first)
  {
    assert_debug(n>=0);
    n_child=n;
    child=first;
  }

  // Return value: number of children
  int NChildren() const  {return n_child;}

  // Parameters: i - index of the child element
  // Return value: id of the i-th child element
  int Child(int i) const
  {
    assert_debug(i>=0 && i<n_child);
    return child+i;
  }

  // Return value: is an active element?
  bool IsActive() const {return (status&ACTIVE)!=0;}

  // Set the element to be active
  void SetActive() {status|=ACTIVE;}

  // Set the element to be inactive
  void UnSetActive() {status&=~ACTIVE;}

  // Set the element to be refined
  void SetToBeRefined() {Status()|=REFINE;}

  // Set the element to be unrefined
  void SetToBeUnRefined() {Status()|=UNREFINE;}

  // Unset the element to be refined
  void UnSetToBeRefined() {Status()&=~REFINE;}

  // Unset the element to be unrefined
  void UnSetToBeUnRefined() {Status()&=~UNREFINE;}

  // Return value: is the element refined?
  bool IsToBeRefined() const {return (status&REFINE)!=0;}

  // Return value: is the element unrefined?
  bool IsToBeUnRefined() const {return (status&UNREFINE)!=0;}

  // Unset all adaptivity switches
  void UnSetAdaptivitySwitches() {Status() &= ~(REFINE|UNREFINE);}  

  // Return value: is the element refined?
  bool IsRefined() const {return (status&SPLIT)!=0;}

  // Return value: level of refinement of the element, rvalue
  unsigned char RefinementLevel() const {return refinement_level;}

  // Return value: level of refinement of the element, rvalue
  unsigned char& RefinementLevel() {return refinement_level;}

  // Refine the triangle using edge bisection
  // Parameters: element_child - child elements; edge_new - new edges
  //             edge_child - child edges; node - nodes on the edge; node_new - new nodes
  // Return value: id of the first child element
  void Refine(TRIANGLE_2D* element_child[], EDGE_2D* edge_new[vert_h], 
              EDGE_2D* edge_child[vert_h][2], NODE_2D* node_new[vert_h]);

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream& out, const TRIANGLE_2D& element);
};

#endif
