//----------------------------------------------------------------------------------------
// mesh_3d.H: Contains declarations for nodes, faces, and prisms in 3D
//----------------------------------------------------------------------------------------

#ifndef MESH_3D_H
#define MESH_3D_H

#include <fstream>
#include <cassert>
#include <vector>
#include "general.H"
#include "map_info.H"
#include "mesh_2d.H"


//--------------------------------------------------------------------------------------
// NODE_3D Class: Node class in 3D
//--------------------------------------------------------------------------------------

class NODE_3D
{
private:
  // System data
  int id;               // Node ID
  int status;           // Status field

  // Connectivity data
  int node_2d;          // Pointer to the correspinding 2D node

  // Geometry data
  real coord[d_space];  // Node coordinates
  real scale;           // Combined volume of elements surrounding the node

public:
  enum status_value {SURFACE=2};  // Status values

  // Default constructor
  NODE_3D(const NODE_2D& node_xy);

  // Constructor for hot start input
  // Parameters: hot start input file stream
  NODE_3D(std::ifstream& in);

  // Default destructor
  ~NODE_3D(){}

  // Return value: Node ID, rvalue
  int ID() const {return id;}

  // Return value: Node ID, lvalue
  int& ID() {return id;}

  // Return value: ID of the corresponding 2D node
  int Node2D() const {return node_2d;}

  // Return value: node coordinates
  const real* Coord() const {return coord;}

  // Parameters: i - coordinate index 
  // Return value: i-th node coordinate, rvalue
  real Coord(int i) const
  {
    assert_debug(i>=0 && i<d_space);
    return coord[i];
  }

  // Parameters: i - coordinate index 
  // Return value: i-th node coordinate, lvalue
  real& Coord(int i)
  {
    assert_debug(i>=0 && i<d_space);
    return coord[i];
  }

  // Return value: the "Z" node coordinate, lvalue
  real ZCoordinate() const {return coord[2];}

  // Return value: the "Z" node coordinate, lvalue
  real& ZCoordinate() {return coord[2];}

  // Combined volume of elements surrounding the node, rvalue
  real Scale() const {return scale;}

  // Combined volume of elements surrounding the node, lvalue
  real& Scale() {return scale;}

  // Return value: Is the surface node?
  bool IsSurface() const {return (status&SURFACE)!=0;}

  // Set the surface flag for the node
  void SetSurface(){status|=SURFACE;}

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream& out, const NODE_3D& node);
};



//--------------------------------------------------------------------------------------
// FACE_HORIZ Class: Horizontal face class in 3D
//--------------------------------------------------------------------------------------

class FACE_HORIZ
{
private:
  // System data
  int id;                // Unique id of the face
  int type;              // Face type field
  int status;            // Status field

  // Geometry data
  MapHorizFace mapping;  // Map from the reference face
  long_real normal[d_space];  // Normal vector 

  // Connectivity data
  int element_2d;        // id of the corresponding 2d face
  int node[vert_h];      // Pointers to face endpoints
  int element[2];        // Pointers to elements sharing the face

  // Hierarchy data
  int parent;            // id of the parent face
  int n_child;           // number of child faces
  int child;             // id of the first child face

public:
  enum status_value {SPLIT=1, ACTIVE=2, SURFACE=4, BOTTOM=8};

  // Default constructor
  FACE_HORIZ(const TRIANGLE_2D& face);

  // Constructor for hot start input
  // Parameters: hot start input file stream
  FACE_HORIZ(std::ifstream& in);

  // Default destructor
  ~FACE_HORIZ() {}

  // Compute mapping from the reference face
  // Parameters: node_coord - node coordinates
  void ComputeReferenceMapping(const real* node_coord[vert_h]);

  // Return value: Mapping from the reference face
  const MapHorizFace& Mapping() const {return mapping;}

  // Return value: Face ID, rvalue
  int ID() const {return id;}

  // Return value: Face ID, lvalue
  int& ID() {return id;}

  // Return value: Face type, rvalue
  int Type() const {return type;}

  // Return value: Face type, lvalue
  int& Type() {return type;}

  // Return value: Status, rvalue
  int Status() const {return status;}

  // Return value: Status, lvalue
  int& Status() {return status;}

  // Return value: Normal vector
  const long_real* Normal() const {return normal;}

  // Parameters: i - coordinate index
  // Return value: i-th normal component, rvalue
  long_real Normal(int i) const
  {
    assert_debug(i>=0 && i<d_space);
    return normal[i];
  }

  // Parameters: i - coordinate index
  // Return value: i-th normal component, lvalue
  long_real& Normal(int i)
  {
    assert_debug(i>=0 && i<d_space);
    return normal[i];
  }

  // Parameters: i - node index
  // Return value: i-th node id, rvalue
  int Node(int i) const
  {
    assert_debug(i>=0 && i<vert_h);
    return node[i];
  }

  // Parameters: i - node index
  // Return value: i-th node id, lvalue
  int& Node(int i)
  {
    assert_debug(i>=0 && i<vert_h);
    return node[i];
  }

  // Parameters: i - element index
  // Return value: i-th element id, rvalue
  int Element(int i) const
  {
    assert_debug(i>=0 && i<2);
    return element[i];
  }

  // Parameters: i - element index
  // Return value: i-th element id, lvalue
  int& Element(int i)
  {
    assert_debug(i>=0 && i<2);
    return element[i];
  }

  // Parameters: i - element index
  // Return value: is i-th element present?
  bool IsElement(int i) const
  {
    assert_debug(i>=0 && i<2);
    return element[i]!=BAD_INDEX;
  }

  // Parameters: el_id - element id
  // Return value: element index if element is contiguous, BAD_INDEX otherwise
  int ElementIndex(int el_id) const
  {int ind = el_id==element[0] ? 0 : (el_id==element[1] ? 1 : BAD_INDEX); return ind;}

  // Return value: id of the corresponding 2D element, rvalue
  int Element2D() const {return element_2d;}

  // Return value: id of the corresponding 2D element, rvalue
  int& Element2D() {return element_2d;}

  // Return value: id of the parent face (if present), rvalue
  int Parent() const {return parent;}

  // Return value: id of the parent face (if present), lvalue
  int& Parent() {return parent;}

  // Parameters: n - number of children
  // Return value: id of the first child face
  void SetChildren(int n, int first)
  {
    assert_debug(n>=0);
    n_child=n;
    child=first;
  }

  // Return value: number of children
  int NChildren() const  {return n_child;}

  // Parameters: i - index of the child face
  // Return value: id of the i-th child face
  int Child(int i) const
  {
    assert_debug(i>=0 && i<n_child);
    return child+i;
  }

  // Return value: is it a surface face?
  bool IsSurface() const {return (status&SURFACE)!=0;}

  // Set the surface face flag
  void SetSurface() {status|=SURFACE;}

  // Return value: is it a bottom face?
  bool IsBottom() const {return (status&BOTTOM)!=0;}

  // Set the bottom face flag
  void SetBottom() {status|=BOTTOM;}

  // Return value: is it an interior face?
  bool IsInterior() const {return (status&BOTTOM || status&SURFACE)==0;}

  // Set the interior face flag
  void SetInterior() {status&=!BOTTOM; status&=!SURFACE;}

  // Return value: is an active face?
  bool IsActive() const {return (status&ACTIVE)!=0;}

  // Set the face to be active
  void SetActive() {status|=ACTIVE;}

  // Set the face to be inactive
  void UnSetActive() {status&=~ACTIVE;}

  // Return value: is the face refined?
  bool IsRefined() const {return (status&SPLIT)!=0;}

  // Refine the face using edge bisection
  // Parameters: element_2d_child - child triangles in 2D; face_horiz_child - child faces;
  //             node - nodes on the face; node_new - new nodes
  // Return value: id of the new node
  void Refine(TRIANGLE_2D *element_2d_child[4], FACE_HORIZ *face_horiz_child[4], 
              NODE_3D *node[vert_h], NODE_3D *node_new[vert_h]);

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream& out, const FACE_HORIZ& face);
};



//--------------------------------------------------------------------------------------
// FACE_VERT Class: Vertical face class in 3D
//--------------------------------------------------------------------------------------

class FACE_VERT
{
private:
  // System data
  int id;                    // Unique id of the face
  int type;                  // Face type field
  int status;                // Status field

  // Geometry data
  MapVertFace mapping;       // Map from the reference face
  long_real normal[d_space]; // Normal vector 

  // Connectivity data
  int edge_2d;               // id of the corresponding 2d face
  int node[4];               // Pointers to face endpoints
  int element[2];            // Pointers to elements sharing the face

  // Hierarchy data
  static std::vector<int> child_vert; // Storage for IDs of vertical split child faces
  int parent;                         // id of the parent face
  int n_child_vert, n_child_horiz;    // number of child faces
  int child_horiz;                    // id of the first horizontal split child face
  int child_vert_ind;                 // index of the id of the first vertical split child face

public:
  enum status_value {HORIZ_SPLIT=1, VERT_SPLIT=2, SURFACE_LAYER=4, ACTIVE=8, INTERIOR=16};

  // Default constructor
  FACE_VERT(const EDGE_2D& edge);

  // Constructor for hot start input
  // Parameters: hot start input file stream
  FACE_VERT(std::ifstream& in);

  // Default destructor
  ~FACE_VERT() {}

  // Compute mapping from the reference face
  // Parameters: node_coord - node coordinates
  void ComputeReferenceMapping(const real* node_coord[4]);

  // Return value: Mapping from the reference face
  const MapVertFace& Mapping() const {return mapping;}

  // Return value: Face ID, rvalue
  int ID() const {return id;}

  // Return value: Face ID, lvalue
  int& ID() {return id;}

  // Return value: Face type, rvalue
  int Type() const {return type;}

  // Return value: Face type, lvalue
  int& Type() {return type;}

  // Return value: Status, rvalue
  int Status() const {return status;}

  // Return value: Status, lvalue
  int& Status() {return status;}

  // Return value: Normal vector
  const long_real* Normal() const {return normal;}

  // Parameters: i - coordinate index
  // Return value: i-th normal component, rvalue
  long_real Normal(int i) const
  {
    assert_debug(i>=0 && i<d_space);
    return normal[i];
  }
  // Parameters: i - coordinate index
  // Return value: i-th normal component, lvalue
  long_real& Normal(int i)
  {
    assert_debug(i>=0 && i<d_space);
    return normal[i];
  }

  // Parameters: i - node index
  // Return value: i-th node id, rvalue
  int Node(int i) const
  {
    assert_debug(i>=0 && i<4);
    return node[i];
  }

  // Parameters: i - node index
  // Return value: i-th node id, lvalue
  int& Node(int i)
  {
    assert_debug(i>=0 && i<4);
    return node[i];
  }

  // Parameters: i - element index
  // Return value: i-th element id, rvalue
  int Element(int i) const
  {
    assert_debug(i>=0 && i<2);
    return element[i];
  }

  // Parameters: i - element index
  // Return value: i-th element id, lvalue
  int& Element(int i)
  {
    assert_debug(i>=0 && i<2);
    return element[i];
  }

  // Parameters: i - element index
  // Return value: is i-th element present?
  bool IsElement(int i) const
  {
    assert_debug(i>=0 && i<2);
    return element[i]!=BAD_INDEX;
  }

  // Parameters: el_id - element id
  // Return value: element index if element is contiguous, BAD_INDEX otherwise
  int ElementIndex(int elem_ind) const
  {int ind = elem_ind==element[0] ? 0 : (elem_ind==element[1] ? 1 : BAD_INDEX); return ind;}

  // Return value: id of the corresponding 2D edge, rvalue
  int Edge2D() const {return edge_2d;}

  // Return value: id of the corresponding 2D edge, lvalue
  int& Edge2D() {return edge_2d;}

  // Return value: id of the parent face (if present), rvalue
  int Parent() const {return parent;}

  // Return value: id of the parent face (if present), lvalue
  int& Parent() {return parent;}

  // Parameters: n - number of children in vertical split; first - id of the first child face
  void SetChildrenVert(int n, int first=BAD_INDEX);

  // Parameters: n - number of children in horizontal split
  // Return value: id of the first child face (if present), lvalue
  void SetChildrenHoriz(int n, int first);

  // Return value: number of children in vertical split
  int NChildrenVert() const  {return n_child_vert;}

  // Return value: number of children in horizontal split
  int NChildrenHoriz() const  {return n_child_horiz;}

  // Parameters: i - index of the child face
  // Return value: id of the i-th child face, rvalue
  int ChildVert(int i) const
  {
    assert_debug(i>=0 && i<n_child_vert);
    return child_vert[child_vert_ind+i];
  }

  // Parameters: i - index of the child face
  // Return value: id of the i-th child face, lvalue
  int& ChildVert(int i)
  {
    assert_debug(i>=0 && i<n_child_vert);
    return child_vert[child_vert_ind+i];
  }

  // Parameters: i - index of the child face
  // Return value: id of the i-th child face, rvalue
  int ChildHoriz(int i) const
  {
    assert_debug(i>=0 && i<n_child_horiz);
    return child_horiz+i;
  }

  // Return value: is it a surface face?
  bool IsSurface() const {return (status&SURFACE_LAYER)!=0;}

  // Set the surface face flag
  void SetSurface() {status|=SURFACE_LAYER;}

  // Return value: is it an interior face?
  bool IsInterior() const {return (status&INTERIOR)!=0;}

  // Set the interior face flag
  void SetInterior() {status|=INTERIOR;}

  // Return value: is the face refined vertically?
  bool IsVertRefined() const {return (status&VERT_SPLIT)!=0;}

  // Return value: is the face refined horizontally?
  bool IsHorizRefined() const {return (status&HORIZ_SPLIT)!=0;}

  // Return value: is an active face?
  bool IsActive() const {return (status&ACTIVE)!=0;}

  // Set the face to be active
  void SetActive() {status|=ACTIVE;}

  // Set the face to be inactive
  void UnSetActive() {status&=~ACTIVE;}

  // Refine the face horizontally
  // Parameters: edge_2d_child - child edges on the 2D edge; face_vert_child - child faces;
  //             node - nodes on the face; node_new - new node
  // Return value: id of the first child face
  void RefineHoriz(EDGE_2D* edge_2d_child[2], FACE_VERT* face_vert_child[2], 
                   NODE_3D* node[4], NODE_3D* node_new[2]);

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream& out, const FACE_VERT& face);

  // Input for storage for IDs of vertical split child faces
  static void InputChildVert(std::ifstream& in)
  {
    std::size_t list_length;
    in.read((char*)&list_length, sizeof list_length);
    child_vert.resize(list_length);
    in.read((char*)&child_vert[0], list_length*sizeof(child_vert[0]));
  } 

  // Output for storage for IDs of vertical split child faces
  static void OutputChildVert(std::ofstream& out)
  {
    std::size_t list_length=child_vert.size();
    out.write((char*)&list_length, sizeof list_length);
    out.write((char*)&child_vert[0], list_length*sizeof(child_vert[0]));
  } 

};



//--------------------------------------------------------------------------------------
// PRISM Class: Element class in 3D
//--------------------------------------------------------------------------------------

class PRISM
{
private:
  // System data
  int id;               // Unique id of the prism
  int type;             // Prism type field
  int status;           // Status field

  // Geometry data
  MapPrism mapping;     // Mapping from the reference prism

  // Connectivity data
  int element_2d;       // id of the face in xy-plane corresponding to the prism
  int top;              // id of the top face
  int bottom;           // id of the bottom face
  int side[vert_h];     // id's of the side faces
  char index[vert_h];   // Element index in contiguous vertical faces

  // Hierarchy data
  int parent;           // id of the parent prism
  int n_child;          // number of child prisms
  int child;            // id of the first child prism

public:
  enum status_value {SPLIT=1, ACTIVE=2, SURFACE=4, BOTTOM=8};

  // Default constructor
  PRISM(const TRIANGLE_2D& element);

  // Constructor for hot start input
  // Parameters: hot start input file stream
  PRISM(std::ifstream& in);

  // Default destructor
  ~PRISM() {}

  // Compute mapping from the reference triangle
  // Parameters: node_coord - node coordinates
  void ComputeReferenceMapping(const real* node_coord[2*vert_h]);

  // Return value: Mapping from the reference triangle
  const MapPrism& Mapping() const {return mapping;}

  // Return value: Element ID, rvalue
  int ID() const {return id;}

  // Return value: Element ID, lvalue
  int& ID() {return id;}

  // Return value: Element type, rvalue
  int Type() const {return type;}

  // Return value: Element type, lvalue
  int& Type() {return type;}

  // Return value: Status, rvalue
  int Status() const {return status;}

  // Return value: Status, lvalue
  int& Status() {return status;}

  // Return value: id of the top face, rvalue
  int TopFace() const {return top;}

  // Return value: id of the top face, lvalue
  int& TopFace() {return top;}

  // Return value: id of the bottom face, rvalue
  int BottomFace() const {return bottom;}

  // Return value: id of the bottom face, lvalue
  int& BottomFace() {return bottom;}

  // Parameters: i - vertical face index
  // Return value: id of the i-th vertical face, rvalue
  int SideFace(int i) const
  {
    assert_debug(i>=0 && i<vert_h);
    return side[i];
  }

  // Parameters: i - vertical face index
  // Return value: id of the i-th vertical face, rvalue
  int& SideFace(int i)
  {
    assert_debug(i>=0 && i<vert_h);
    return side[i];
  }

  // Parameters: face - vertical face id
  // Return value: index of the vertical face in the element
  int FaceVertIndex(int face) const
  {for (int i=0; i<vert_h; i++) if (side[i]==face) return i; return BAD_INDEX;}

  // Parameters: i - vertical face index
  // Return value: element index in ith vertical face, rvalue
  char ElementIndex(int i) const
  {
    assert_debug(i>=0 && i<vert_h);
    return index[i];
  }

  // Parameters: i - vertical face index
  // Return value: element index in ith vertical face, lvalue
  char& ElementIndex(int i)
  {
    assert_debug(i>=0 && i<vert_h);
    return index[i];
  }

  // Return value: id of the corresponding 2D element, rvalue
  int Element2D() const {return element_2d;}

  // Return value: id of the corresponding 2D element, lvalue
  int& Element2D() {return element_2d;}

  // Return value: id of the parent element (if present), rvalue
  int Parent() const {return parent;}

  // Return value: id of the parent element (if present), lvalue
  int& Parent() {return parent;}

  // Parameters: n - number of children
  // Return value: id of the first child element
  void SetChildren(int n, int first)
  {
    assert_debug(n>=0);
    n_child=n;
    child=first;
  }

  // Return value: number of children
  int NChildren() const  {return n_child;}

  // Parameters: i - index of the child element
  // Return value: id of the i-th child element
  int Child(int i) const
  {
    assert_debug(i>=0 && i<n_child);
    return child+i;
  }

  // Return value: is it a surface face?
  bool IsSurface() const {return (status&SURFACE)!=0;}

  // Set the surface face flag
  void SetSurface() {status|=SURFACE;}

  // Return value: is it a bottom face?
  bool IsBottom() const {return (status&BOTTOM)!=0;}

  // Set the bottom face flag
  void SetBottom() {status|=BOTTOM;}

  // Return value: is an active element?
  bool IsActive() const {return (status&ACTIVE)!=0;}

  // Set the element to be active
  void SetActive() {status|=ACTIVE;}

  // Set the element to be inactive
  void UnSetActive() {status&=~ACTIVE;}  

  // Return value: is the element refined?
  bool IsRefined() const {return (status&SPLIT)!=0;}

  // Refine the element
  // Parameters: element_2d_child - 2D child elements; edge_2d_new - new 2D edges
  //             prism_child - child elements; face_vert_new - new vertical faces;
  //             face_vert_child - child vertical faces;
  //             face_horiz_child_top - child horizontal faces on the top face;
  //             face_horiz_child_bot - child horizontal faces on the bottom face;
  //             node_new - new nodes
  // Return value: id of the first child element
  void Refine(TRIANGLE_2D* element_2d_child[4], EDGE_2D* edge_2d_new[vert_h],
              PRISM* prism_child[4], FACE_VERT* face_vert_new[vert_h],
              FACE_VERT* face_vert_child[vert_h][2], FACE_HORIZ* face_horiz_child_top[4],
              FACE_HORIZ* face_horiz_child_bot[4], NODE_3D* node_new[2][vert_h]);

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream& out, const PRISM& prism);
};

#endif
