#include <iostream>
#include <cmath>
#include "general.H"
#include "pde.H"
#include "storage_container.H"
#include "mesh_2d_sw.H"
#include "mesh_3d_sw.H"

extern PROBLEM_DESCRIPTION problem_info;
#if TURBULENCE_KEY==3  // two equation turbulence models
extern GLS_Data gls_data;     // Data for two equation turbulence models
#endif



//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
//                                  PDE functions
//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

// Normal (upwind) flux for the transport term
// Parameters: var_l, var_r - values of the unknowns on both sides of the face; flux - result
inline
void transport_flux(const real var_l[d_var], const real var_r[d_var], long_real flux[d_var])
{
  static const VAR* ind=VAR_TRANSPORT_INDEX;
  const real *var = flux[XI]>0. ? var_l : var_r;

  for (int i=0; i<d_var_transport; i++)
    flux[ind[i]]=flux[XI]*var[ind[i]];

//   for (int i=0; i<d_var_transport; i++)
//     flux[ind[i]]=flux[XI]*0.5*(var_l[ind[i]]+var_r[ind[i]]);
}


// Normal advective flux on horizontal faces
// Parameters: normal - normal to the face; var - values of the unknowns; flux - result
inline
void advection_horiz_flux(const long_real normal[d_space], const real var[d_var], long_real flux[d_var])
{
  long_real advec[d_var][d_space];

  advection_horiz(var, advec);
  for (int i=0; i<d_space; i++)
    {
      flux[XI]+=advec[XI][i]*normal[i];
      flux[U]+=advec[U][i]*normal[i];
      flux[V]+=advec[V][i]*normal[i];
    }
}


// Riemann solver on vertical faces
// Parameters: TClass - face/element pointer; 
//             var_l, var_r - values of the unknowns on both sides of the face; flux - result
inline
void RiemannSolver(const FACE_VERT_SW *this_obj, const real var_l[d_var], const real var_r[d_var],
                   long_real flux[d_var])
{
  switch(this_obj->Type())
  {
// #ifndef SUPERCRITICAL_KEY
//   case EDGE_2D_SW::LAND:
// #endif
//   case EDGE_2D_SW::SEA: 
//   case EDGE_2D_SW::RIVER:
  case EDGE_2D_SW::RADIATION:
    advection_vert_flux(this_obj->Normal(), var_l, var_r, flux);
    break;

  default:
    switch (problem_info.Riemann_vert_switch)
    {
    case 0: Roe(this_obj->Normal(), var_l, var_r, flux);
      break;
    case 1: RoeEntropyFix(this_obj->Normal(), var_l, var_r, flux);
      break;
    case 2: HLL(this_obj->Normal(), var_l, var_r, flux);
      break;
    case 3: HLLC(this_obj->Normal(), var_l, var_r, flux);
      break;
    case 4: FORCE(this_obj->Normal(), var_l, var_r, flux);
      break;
    case 5: LaxFriedrichs(this_obj->Normal(), var_l, var_r, flux);
      break;
    default: std::cerr << "File numerics.cpp: Riemann solver choice parameter out of range!\n";
      std::exit(1);
    }
  }
}


//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
//                       Member functions for numerics
//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------
//                      Projection computation functions
//--------------------------------------------------------------------------------------

void TRIANGLE_2D_SW::Projection_elevation(real t, void func(real, const real[], real[]))
{
  real elevation;

  func(t, coord_quad(), &elevation);
  Var_dof(XI).Add(weight_quad()*elevation, basis_func_quad(XI));
}

void TRIANGLE_2D_SW::Projection_elevation_forcing(real t, void func(real, const real[], real[]))
{
  real elevation_forcing;

  func(t, coord_quad(), &elevation_forcing);
  L_Var_dof(XI).Add(weight_quad()*elevation_forcing, basis_func_quad(XI));
}

void PRISM_SW::Projection(real t, void func(real, const real[], real[]))
{
  static const VAR* ind=VAR_3D_DT_INDEX;
  real var[d_var]={0.};

  func(t, coord_quad(), var);
  for (int i=0; i<d_var_3d_dt; i++)
    Var_dof(ind[i]).Add(weight_quad()*var[ind[i]], basis_func_quad(ind[i]));
}

void PRISM_SW::Projection_forcing(real t, void func(real, const real[], real[]))
{
  static const VAR* ind=VAR_3D_DT_INDEX;
  real var[d_var]={0.};

  func(t, coord_quad(), var);
  for (int i=0; i<d_var_3d_dt; i++)
    L_Var_dof(ind[i]).Add(weight_quad()*var[ind[i]], basis_func_quad(ind[i]));
}

#ifdef MASS_CONSERVATION_KEY
void PRISM_SW::Tracer_Projection(real t)
{
  real source_term = Element2D()==1378 && t>=1000. && t<=2000. ? .0001 : 0.;
  L_Var_dof(TRACER).Add(weight_quad()*source_term, basis_func_quad(TRACER));
  mass_change()+=weight_quad()*scale()*source_term;
}

void PRISM_SW::Tracer_Mass()
{
  norm(VAR_3D_DT_INDEX_INV[TRACER])+=weight_quad()*scale()*(Var_dof(TRACER)*basis_func_quad(TRACER));
}
#endif

//--------------------------------------------------------------------------------------
//                    Error and time change computation functions
//--------------------------------------------------------------------------------------

void TRIANGLE_2D_SW::Elevation_change_L2()
{
  long_real d_xi=L_Var_dof(XI)*basis_func_quad(XI);
  norm()+=weight_quad()*scale()*d_xi*d_xi;
}

void TRIANGLE_2D_SW::Elevation_error_L2(real t, void func(real, const real[], real[]))
{
  real elevation, error_xi;

  func(t, coord_quad(), &elevation);
  error_xi=Var_quad(XI)-elevation;
  norm()+=weight_quad()*scale()*error_xi*error_xi;
}

void PRISM_SW::Global_change_L2()
{
  static const VAR* ind=VAR_3D_DT_INDEX;
  for (int i=0; i<d_var_3d_dt; i++)
  {
    long_real d_u=L_Var_dof(ind[i])*basis_func_quad(ind[i]);
    norm(i)+=weight_quad()*scale()*d_u*d_u;
  }
}

void PRISM_SW::Global_error_L2(real t, void func(real, const real[], real[]))
{
  static const VAR* ind=VAR_3D_INDEX;
  real var[d_var]={0.};

  func(t, coord_quad(), var);
  for (int i=0; i<d_var_3d; i++)
  {
    real error_u=Var_dof(ind[i])*basis_func_quad(ind[i])-var[ind[i]];
    norm(i)+=weight_quad()*scale()*error_u*error_u;
  }
}

// Compute depth averaged horizontal velocity components
#ifdef DEPTH_AVERAGED_VELOCITY_KEY
void PRISM_SW::ComputeDepthAveragedVelocity(real t)
{
  real vel[d_space_2d]={Var_dof(U)*basis_func_quad(U), Var_dof(V)*basis_func_quad(V)};
  real coef=1.;

    Var_dof_2D(UH).Add(elev_weight_quad()*coef*vel[0], basis_func_quad_2D(UH));
    Var_dof_2D(VH).Add(elev_weight_quad()*coef*vel[1], basis_func_quad_2D(VH));
}
#endif

//--------------------------------------------------------------------------------------
//                     Advection computation functions
//--------------------------------------------------------------------------------------

//---------------------------------------
void PRISM_SW::ElementIntegration(real t)
//---------------------------------------
{
  static const VAR *ind=VAR_3D_DT_INDEX, *ind_t=VAR_TRANSPORT_INDEX;
  real var[d_var]={0.}, var_c[d_var]={0.}, rhs[d_var]={0.};
  long_real advec[d_var][d_space]={{0.}}, advec_rf[d_var][d_space]={{0.}};

  // evaluate functions at quadrature points
  var[XI]=Var_2D_quad(XI)+depth_quad();
  for (int i=0; i<d_var_3d_dt; i++)
    var[ind[i]]=Var_dof(ind[i])*basis_func_quad(ind[i]);
  var[W]=Var_dof(W)*basis_func_quad(W);

  // Compute advection and transport terms
  advection(var, advec);
  transport(var, advec);
  advec[U][0]-=problem_info.g*depth_quad();
  advec[V][1]-=problem_info.g*depth_quad();
#ifdef BAROCLINIC_KEY
  real density_forcing=problem_info.g/problem_info.Rho_0*density_forcing_quad();
  advec[U][0]+=density_forcing;
  advec[V][1]+=density_forcing;
#endif
#ifdef ROUNDOFF_KEY  // Roundoff error reduction treatment
  var_c[U]=var[U]; var_c[V]=var[V]; var_c[W]=var[W];
  for (int i=0; i<d_var_transport; i++)
    var_c[ind_t[i]]=Var_dof(ind_t[i])(0)*basis_func_quad(ind_t[i])[0];  
  transport(var_c, advec_rf);
  var_c[U]=var_c[V]=var_c[W]=0.;
#endif
  rhs[U]=problem_info.g*tidal_forcing(0)+Coriolis_quad()*var[V];
  rhs[V]=problem_info.g*tidal_forcing(1)-Coriolis_quad()*var[U];
  for (int i=0; i<d_var_3d_dt; i++)
  {
    // Add advection and transport terms
    for (int j=0; j<d_space; j++)
      L_Var_dof(ind[i]).
        Add(weight_quad()*(advec[ind[i]][j]-advec_rf[ind[i]][j]), basis_grad_quad(ind[i], j));
    // Add body forces
    L_Var_dof(ind[i]).Add(weight_quad()*rhs[ind[i]], basis_func_quad(ind[i]));
  }
  // Add contribution to the surface elevation
  for (int i=0; i<d_space_2d; i++) 
    L_Var_dof_2D(XI).Add(elev_weight_quad()*advec[XI][i], basis_grad_quad_2D(XI, i));

  // Compute Q-variables (gradients of the primary variables)
  // Vertical components of Q
  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind_diff=problem_info.diff_vert_ind[i];
    Q_dof(ind_diff, 2).Add(weight_quad()*(var[ind_diff]-var_c[ind_diff]), basis_grad_quad(ind_diff, 2));
  }
  // Horizontal components of Q
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind_diff=problem_info.diff_horiz_ind[i];
    for (int j=0; j<d_space_2d; j++)
      Q_dof(ind_diff, j).Add(weight_quad()*(var[ind_diff]-var_c[ind_diff]), basis_grad_quad(ind_diff, j));
  }
}

//--------------------------------------- 
void FACE_HORIZ_SW::FaceHorizIntegration(real t)
//---------------------------------------
{
  static const VAR *ind=VAR_3D_DT_INDEX, *ind_t=VAR_TRANSPORT_INDEX;
  real var[d_var]={0.};
  long_real f_hat[d_var]={0.};

  // evaluate functions at quadrature points
  var[XI]=Var_2D_quad(XI)+depth_quad();
  if (IsInterior())
  {
    for (int i=0; i<d_var_3d_dt; i++)
      var[ind[i]]=0.5*(Var_dof(0, ind[i])*basis_func_quad(0, ind[i])+
                       Var_dof(1, ind[i])*basis_func_quad(1, ind[i]));
    var[W]=Var_dof(1, W)*basis_func_quad(1, W);
  }
  else
    BC_Dirichlet_quad(t, 0, var);

  // Compute advection and transport terms
  advection_horiz_flux(Normal(), var, f_hat);
  transport_flux(var, var, f_hat);
  f_hat[U]-=problem_info.g*depth_quad()*Normal(0);
  f_hat[V]-=problem_info.g*depth_quad()*Normal(1);
#ifdef BAROCLINIC_KEY
  real density_forcing=problem_info.g/problem_info.Rho_0*density_forcing_quad();
  f_hat[U]+=density_forcing*Normal(0);
  f_hat[V]+=density_forcing*Normal(1);
#endif
  for (int k=0; k<2; k++)
    if (IsElement(k))
    {
      real var_c[d_var]={0.};
      long_real f_rf[d_var]={0.};
#ifdef ROUNDOFF_KEY  // Roundoff error reduction treatment
      var_c[U]=var[U]; var_c[V]=var[V]; var_c[W]=var[W];
      f_rf[XI]=f_hat[XI];
      for (int i=0; i<d_var_transport; i++)
        var_c[ind_t[i]]=Var_dof(k, ind_t[i])(0)*basis_func_quad(k, ind_t[i])[0];
      transport_flux(var_c, var_c, f_rf);
      var_c[U]=var_c[V]=var_c[W]=0.;
#endif
      // Add advection and transport terms
      for (int i=0; i<d_var_3d_dt; i++)
        L_Var_dof(k, ind[i]).
          Add(-coef_scale(k)*weight_quad()*(f_hat[ind[i]]-f_rf[ind[i]]), basis_func_quad(k, ind[i]));
      // Compute Q-variables (gradients of the primary variables)
      // Vertical components of Q
      for (int i=0; i<problem_info.n_diff_vert; i++)
      {
        VAR ind_diff=problem_info.diff_vert_ind[i];
        Q_dof(k, ind_diff, 2).
          Add(-coef_scale(k)*weight_quad()*(var[ind_diff]-var_c[ind_diff])*Normal(2),
              basis_func_quad(k, ind_diff));
      }
      // Horizontal components of Q
      for (int i=0; i<problem_info.n_diff_horiz; i++)
      {
        VAR ind_diff=problem_info.diff_horiz_ind[i];
        for (int j=0; j<d_space_2d; j++)
          Q_dof(k, ind_diff, j).
            Add(-coef_scale(k)*weight_quad()*(var[ind_diff]-var_c[ind_diff])*Normal(j), 
                basis_func_quad(k, ind_diff));
      }
    }
#ifdef TEST_KEY
    if (IsBottom()) // Contribution to mass conservation from flow through the bottom faces
      L_Var_dof_2D(XI).Add(weight_quad()*f_hat[XI]/Normal(2), basis_func_quad_2D(XI));
#endif
    if (!IsSurface()) // Continuity equation term on all except surface faces
      Var_dof(0, W).Add(-coef_scale(0)*weight_quad()*f_hat[XI], basis_func_quad(0, W));
}

//------------------------------------- 
void FACE_VERT_SW::FaceVertIntegration(real t)
//-------------------------------------
{
  static const VAR *ind=VAR_3D_DT_INDEX, *ind_t=VAR_TRANSPORT_INDEX;
  real var[2][d_var]={{0.}};
  long_real f_hat[d_var]={0.};

  // Evaluate functions at quadrature points
  for (int k=0; k<2; k++)
  {
    var[k][XI]=Var_2D_quad(k, XI)+depth_quad();
    if (IsElement(k))
      for (int i=0; i<d_var_3d_dt; i++)
        var[k][ind[i]]=Var_dof(k, ind[i])*basis_func_quad(k, ind[i]);
    else
      BC_Dirichlet_quad(t, 0, var[k]);
    }

  // Compute advection and transport terms
  RiemannSolver(this, var[0], var[1], f_hat);
  transport_flux(var[0], var[1], f_hat);
  f_hat[U]-=problem_info.g*depth_quad()*Normal(0);
  f_hat[V]-=problem_info.g*depth_quad()*Normal(1);
#ifdef BAROCLINIC_KEY
  real density_forcing=problem_info.g/problem_info.Rho_0*density_forcing_quad();
  f_hat[U]+=density_forcing*Normal(0);
  f_hat[V]+=density_forcing*Normal(1);
#endif
  for (int k=0; k<2; k++)
    if (IsElement(k))
    {
      real var_c[d_var]={0.};
      long_real f_rf[d_var]={0.};
#ifdef ROUNDOFF_KEY  // Roundoff reduction treatment
      var_c[U]=var[k][U]; var_c[V]=var[k][V]; var_c[W]=var[k][W];
      f_rf[XI]=f_hat[XI];
      for (int i=0; i<d_var_transport; i++)
        var_c[ind_t[i]]=Var_dof(k, ind_t[i])(0)*basis_func_quad(k, ind_t[i])[0];
      transport_flux(var_c, var_c, f_rf);
      var_c[U]=var_c[V]=var_c[W]=0.;
#endif
      // Add advection terms
      L_Var_dof_2D(k, XI).
        Add(-coef_scale_elev(k)*elev_weight_quad()*f_hat[XI], basis_func_quad_2D(k, XI));
      L_Var_dof(k, U).
        Add(-coef_scale(k)*weight_quad()*(f_hat[U]-f_rf[U]), basis_func_quad(k, U));
      L_Var_dof(k, V).
        Add(-coef_scale(k)*weight_quad()*(f_hat[V]-f_rf[V]), basis_func_quad(k, V));

      // Add transport terms
      for (int i=0; i<d_var_transport; i++)
        L_Var_dof(k, ind_t[i]).
          Add(-coef_scale(k)*elev_weight_quad()*(f_hat[ind_t[i]]-f_rf[ind_t[i]]), 
              basis_func_quad(k, ind_t[i]));

      // Add continuity equation term
#ifdef SUPERCRITICAL_KEY
      long_real u_n=0.5*((var[0][U]+var[1][U])*Normal(0)+(var[0][V]+var[1][V])*Normal(1));
      Var_dof(k, W).Add(-coef_scale(k)*weight_quad()*u_n, basis_func_quad(k, W));
#else
      Var_dof(k, W).Add(-coef_scale(k)*elev_weight_quad()*f_hat[XI], basis_func_quad(k, W));
#endif
      // Compute Q-variables (gradients of the primary variables)
     for (int i=0; i<problem_info.n_diff_horiz; i++)
      {
        VAR ind_diff=problem_info.diff_horiz_ind[i];
        long_real tmp=-coef_scale(k)*weight_quad()*
          (0.5*(var[0][ind_diff]+var[1][ind_diff])-var_c[ind_diff]);
        for (int j=0; j<d_space_2d; j++)
          Q_dof(k, ind_diff, j).Add(tmp*Normal(j),basis_func_quad(k, ind_diff));
      }
    }
#ifdef MASS_CONSERVATION_KEY
  if (!IsElement(1))
    PRISM_SW::mass_change()-=weight_quad()*f_hat[TRACER];
#endif

  // Adaptivity procedure
#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
    if (IsInterior())
    {
      int ind_adapt = f_hat[XI]>0. ? 1 : 0;
#ifdef ADAPTIVITY_H_2D_KEY
      long_real weight=elev_weight_quad(),
        jump=Var_2D_quad(ind_adapt, problem_info.refinement_var)-
             Var_2D_quad(1-ind_adapt, problem_info.refinement_var);
#endif
#ifdef ADAPTIVITY_H_3D_KEY
      long_real weight=weight_quad(), jump=var[ind_adapt][problem_info.refinement_var]-
        var[1-ind_adapt][problem_info.refinement_var];
#endif
      // 0 - inflow, 1 - outflow
      for (int k=0; k<2; k++)
      {
        JumpBoundary(ind_adapt, 0)+=weight;
        JumpBoundary(1-ind_adapt, 1)+=weight;
        Jump(ind_adapt, 0)+=weight*jump;
        Jump(1-ind_adapt, 1)-=weight*jump;
      }
    }
#endif
}

//--------------------------------------------------------------------------------------
//                       Diffusion computation functions
//--------------------------------------------------------------------------------------

//-----------------------------------------
void PRISM_SW::ElementIntegrationDiffusion(real t)
//-----------------------------------------
{
  real diff_coef[d_var][d_space], param[7]={0.}, rhs[d_var]={0.}, coord_ref[d_space];

  // Compute diffusion term
  switch(problem_info.diff_vert_switch)
  {
  case 0: // No verical eddy viscosity
    break;
  case 1: // Constant vertical eddy viscosity coeffficient
    diffusion_vert_coef_const(param, diff_coef);
    break;
#ifdef DEPTH_AVERAGED_VELOCITY_KEY
  case 2: // Eddy viscosity coef. ~ U*H (U - depth averaged velocity, Davies)
    param[0]=depth_quad();
    param[1]=UH_quad(0);
    param[2]=UH_quad(1);
    diffusion_vert_coef_uh(param, diff_coef);
    break;
  case 3: // Eddy viscosity coefficient ~ U*U (U - depth averaged velocity, Davies)
    param[0]=depth_quad();
    param[1]=UH_quad(0);
    param[2]=UH_quad(1);
    diffusion_vert_coef_uu(param, diff_coef);
    break;
#endif
#if TURBULENCE_KEY>1  // one equation turbulence models
  case 4: // Delft3D
    param[0]=surface_elev_quad()-coord_quad(2);
    param[1]=depth_quad()+coord_quad(2)+problem_info.bottom_l;
#ifdef BAROCLINIC_KEY
    param[2]=-problem_info.g/problem_info.Rho_0*(RHO_Z_dof()*basis_func_quad(RHO));
#else
    param[2]=0.;
#endif
    if (IsBottom() && problem_info.bottom_rough_coef!=0.)
      {
        real coef=param[1]*std::log(param[1]/problem_info.bottom_rough_coef);
        param[3]=(Var_dof(U)*basis_func_quad(U))/coef;
        param[4]=(Var_dof(V)*basis_func_quad(V))/coef;
      }
    else 
      if (IsSurface() && problem_info.surface_bc!=0 && param[1]>param[0])
        {
          const real kappa=0.4;

          param[3]=std::sqrt(std::abs(problem_info.surface_stress[0]))/kappa/param[0];
          param[4]=std::sqrt(std::abs(problem_info.surface_stress[1]))/kappa/param[0];
        }
      else
        {
          param[3]=-(Q_dof(U, 2)*basis_func_quad(U));
          param[4]=-(Q_dof(V, 2)*basis_func_quad(V));
        }
    param[5]=Var_dof(K)*basis_func_quad(K);
    diffusion_vert_coef_k(param, diff_coef, rhs);

    // Add the right hand side contribution for the k model
//     rhs_k(param, rhs);
    L_Var_dof(K).Add(weight_quad()*rhs[K], basis_func_quad(K));
    break;
#if TURBULENCE_KEY==3  // two equation turbulence models
  case 5: // Mellor-Yamada k-kl turbulence model
  case 6: // Burchard et al. k-eps turbulence model
  case 7: // Wilcox et al. k-omega turbulence model
  case 8: // Umlauf-Burchard, 2003, gen model
    param[0]=surface_elev_quad()-coord_quad(2);
    param[1]=depth_quad()+coord_quad(2)+problem_info.bottom_l;
#ifdef BAROCLINIC_KEY
    param[2]=-problem_info.g/problem_info.Rho_0*(RHO_Z_dof()*basis_func_quad(RHO));
#else
    param[2]=0.;
#endif
    if (IsBottom() && problem_info.bottom_rough_coef!=0.)
      {
        real coef=param[1]*std::log(param[1]/problem_info.bottom_rough_coef);
        param[3]=(Var_dof(U)*basis_func_quad(U))/coef;
        param[4]=(Var_dof(V)*basis_func_quad(V))/coef;
      }
    else 
      if (IsSurface() && problem_info.surface_bc!=0 && param[1]>param[0])
        {
          param[3]=std::sqrt(std::abs(problem_info.surface_stress[0]))/gls_data.kappa/param[0];
          param[4]=std::sqrt(std::abs(problem_info.surface_stress[1]))/gls_data.kappa/param[0];
        }
      else
        {
          param[3]=-(Q_dof(U, 2)*basis_func_quad(U));
          param[4]=-(Q_dof(V, 2)*basis_func_quad(V));
        }
    param[5]=Var_dof(K)*basis_func_quad(K);
    param[6]=Var_dof(E)*basis_func_quad(E);
    diffusion_vert_coef_gls(param, diff_coef, rhs);

    // Add the right hand side contribution
//     rhs_gls(param, rhs);
    L_Var_dof(K).Add(weight_quad()*rhs[K], basis_func_quad(K));
    L_Var_dof(E).Add(weight_quad()*rhs[E], basis_func_quad(E));
    break;
#endif
#endif
  default: std::cerr << "File numerics.cpp: Unknown vertical eddy viscosity model choice parameter!\n";
    std::exit(1);
  }
  diffusion_horiz_coef(param, diff_coef);

  // Add contribution to the right hand side
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind=problem_info.diff_horiz_ind[i];
    L_Var_dof(ind).Add(weight_quad()*diff_coef[ind][0]*(Q_dof(ind, 0)*basis_func_quad(ind)), 
                       basis_grad_quad(ind, 0));
    L_Var_dof(ind).Add(weight_quad()*diff_coef[ind][1]*(Q_dof(ind, 1)*basis_func_quad(ind)),
                       basis_grad_quad(ind, 1));
  }
  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind=problem_info.diff_vert_ind[i];
    L_Var_dof(ind).Add(weight_quad()*diff_coef[ind][2]*(Q_dof(ind, 2)*basis_func_quad(ind)),
                       basis_grad_quad(ind, 2));
  }
}

//-----------------------------------------
void FACE_HORIZ_SW::FaceHorizIntegrationDiffusion(real t)
//-----------------------------------------
{
  real diff_coef[d_var][d_space], diff[d_var][d_space]={{0.}}, param[7]={0.};

  // Compute diffusion coefficients
  switch(problem_info.diff_vert_switch)
  {
  case 0: // No verical eddy viscosity
    break;
  case 1: // Constant vertical eddy viscosity coeffficient
    diffusion_vert_coef_const(param, diff_coef);
    break;
#ifdef DEPTH_AVERAGED_VELOCITY_KEY
  case 2: // Eddy viscosity coef. ~ U*H (U - depth averaged velocity, Davies)
    param[0]=depth_quad();
    param[1]=UH_quad(0);
    param[2]=UH_quad(1);
    diffusion_vert_coef_uh(param, diff_coef);
    break;
  case 3: // Eddy viscosity coefficient ~ U*U (U - depth averaged velocity, Davies)
    param[0]=depth_quad();
    param[1]=UH_quad(0);
    param[2]=UH_quad(1);
    diffusion_vert_coef_uu(param, diff_coef);
    break;
#endif
#if TURBULENCE_KEY>1  // one equation turbulence models
  case 4: // Delft3D
    {
      real coef = IsInterior() ? 0.5 : 1.;

      param[0]=surface_elev_quad()-coord_quad(2);
      param[1]=depth_quad()+coord_quad(2)+problem_info.bottom_l;
      param[2]=0.;
      for (int k=0; k<2; k++)
        if (IsElement(k))
          {
#ifdef BAROCLINIC_KEY
            param[2]+=-coef*problem_info.g/problem_info.Rho_0*(RHO_Z_dof(k)*basis_func_quad(k, RHO_Z));
            param[3]+=-coef*((Q_dof(k, U, 2)*basis_func_quad(k, U)));
            param[4]+=-coef*((Q_dof(k, V, 2)*basis_func_quad(k, V)));
#endif
            param[5]+=coef*(Var_dof(k, K)*basis_func_quad(k, K));
          }
    }
//     param[0]=surface_elev_quad()-coord_quad(2);
//     param[1]=depth_quad()+coord_quad(2)+problem_info.bottom_l;
//     param[2]=0.;
//     if (IsInterior())
//     {
// #ifdef BAROCLINIC_KEY
//       param[2]=-problem_info.g/problem_info.Rho_0*
//         0.5*(RHO_Z_dof(0)*basis_func_quad(0, RHO_Z)+RHO_Z_dof(1)*basis_func_quad(1, RHO_Z));
//       param[3]=-0.5*((Q_dof(0, U, 2)*basis_func_quad(0, U))+(Q_dof(1, U, 2)*basis_func_quad(1, U)));
//       param[4]=-0.5*((Q_dof(0, V, 2)*basis_func_quad(0, V))+(Q_dof(1, V, 2)*basis_func_quad(1, V)));
// #endif
//       param[5]=0.5*(Var_dof(0, K)*basis_func_quad(0, K)+Var_dof(1, K)*basis_func_quad(1, K));
//     }
//     else
//     {
//       real var[d_var];

// #ifdef BAROCLINIC_KEY
//       if (IsSurface())
//         {
//           param[2]=-problem_info.g/problem_info.Rho_0*(RHO_Z_dof(1)*basis_func_quad(1, RHO_Z));
//           param[3]=-(Q_dof(1, U, 2)*basis_func_quad(1, U));
//           param[4]=(Q_dof(1, V, 2)*basis_func_quad(1, V));
//         }
//       else
//         {
//           param[2]=-problem_info.g/problem_info.Rho_0*(RHO_Z_dof(0)*basis_func_quad(0, RHO_Z));
//           param[3]=(Q_dof(0, U, 2)*basis_func_quad(0, U));
//           param[4]=(Q_dof(0, V, 2)*basis_func_quad(0, V));
//         }
// #endif
//       BC_Dirichlet_quad(t, 0, var);
//       param[5]=var[K];
//     }
    diffusion_vert_coef_k(param, diff_coef);
    break;
#if TURBULENCE_KEY==3  // two equations turbulence models
  case 5: // Mellor-Yamada k-kl turbulence model
  case 6: // Burchard et al. k-eps turbulence model
  case 7: // Wilcox et al. k-omega turbulence model
  case 8: // Umlauf-Burchard, 2003, gen model
    {
      real coef = IsInterior() ? 0.5 : 1.;

      param[0]=surface_elev_quad()-coord_quad(2);
      param[1]=depth_quad()+coord_quad(2)+problem_info.bottom_l;
      param[2]=0.;
      for (int k=0; k<2; k++)
        if (IsElement(k))
          {
#ifdef BAROCLINIC_KEY
            param[2]+=-coef*problem_info.g/problem_info.Rho_0*(RHO_Z_dof(k)*basis_func_quad(k, RHO_Z));
#endif
            param[5]+=coef*(Var_dof(k, K)*basis_func_quad(k, K));
            param[6]+=coef*(Var_dof(k, E)*basis_func_quad(k, E));
          }
    }

//     if (IsInterior())
//     {
// #ifdef BAROCLINIC_KEY
//       param[2]=-problem_info.g/problem_info.Rho_0*
//         0.5*(RHO_Z_dof(0)*basis_func_quad(0, RHO_Z)+RHO_Z_dof(1)*basis_func_quad(1, RHO_Z));
// #endif
//       param[5]=0.5*(Var_dof(0, K)*basis_func_quad(0, K)+Var_dof(1, K)*basis_func_quad(1, K));
//       param[6]=0.5*(Var_dof(0, E)*basis_func_quad(0, E)+Var_dof(1, E)*basis_func_quad(1, E));
//     }
//     else
//     {
//       real var[d_var];

// #ifdef BAROCLINIC_KEY
//       if (IsSurface())
//         param[2]=-problem_info.g/problem_info.Rho_0*(RHO_Z_dof(1)*basis_func_quad(1, RHO_Z));
//       else
//         param[2]=-problem_info.g/problem_info.Rho_0*(RHO_Z_dof(0)*basis_func_quad(0, RHO_Z));
// #endif
// //       BC_Dirichlet_quad(t, 0, var);
// //       param[5]=var[K];
// //       param[6]=var[E];
//       if (IsSurface())
//         {
//           param[5]=Var_dof(1, K)*basis_func_quad(1, K);
//           param[6]=Var_dof(1, E)*basis_func_quad(1, E);
//         }
//       else
//         {
//           param[5]=Var_dof(0, K)*basis_func_quad(0, K);
//           param[6]=Var_dof(0, E)*basis_func_quad(0, E);
//         }
//     }
    diffusion_vert_coef_gls(param, diff_coef);
    break;
#endif
#endif
    default: std::cerr << "File numerics.cpp: Unknown vertical eddy viscosity model choice parameter!\n";
      std::exit(1);
  }
  diffusion_horiz_coef(param, diff_coef);

  // Compute the diffusion term
  if (IsInterior())
    {
    for (int i=0; i<problem_info.n_diff_horiz; i++)
      {
        VAR ind=problem_info.diff_horiz_ind[i];
        diff[ind][0]=diff_coef[ind][0]*0.5*(Q_dof(0, ind, 0)*basis_func_quad(0, ind)+
                                            Q_dof(1, ind, 0)*basis_func_quad(1, ind));
        diff[ind][1]=diff_coef[ind][1]*0.5*(Q_dof(0, ind, 1)*basis_func_quad(0, ind)+
                                            Q_dof(1, ind, 1)*basis_func_quad(1, ind));
      }
    for (int i=0; i<problem_info.n_diff_vert; i++)
      {
        VAR ind=problem_info.diff_vert_ind[i];
        diff[ind][2]=diff_coef[ind][2]*0.5*(Q_dof(0, ind, 2)*basis_func_quad(0, ind)+
                                            Q_dof(1, ind, 2)*basis_func_quad(1, ind));
      }
    }
  else
    BC_Neumann_quad(t, diff_coef, diff);

  // Add contribution to the right hand side
  for (int k=0; k<2; k++)
    if (IsElement(k))
    {
      for (int i=0; i<problem_info.n_diff_horiz; i++)
      {
        VAR ind=problem_info.diff_horiz_ind[i];
        L_Var_dof(k, ind).
          Add(-coef_scale(k)*weight_quad()*(diff[ind][0]*Normal(0)+diff[ind][1]*Normal(1)), 
              basis_func_quad(k, ind));
      }
      for (int i=0; i<problem_info.n_diff_vert; i++)
      {
        VAR ind=problem_info.diff_vert_ind[i];
        L_Var_dof(k, ind).
          Add(-coef_scale(k)*weight_quad()*diff[ind][2]*Normal(2), basis_func_quad(k, ind));
      }
    }
}

//-----------------------------------------
void FACE_VERT_SW::FaceVertIntegrationDiffusion(real t)
//-----------------------------------------
{
  real diff_coef[d_var][d_space], diff[d_var][d_space]={{0.}}, param[3]={0.};

  // Compute diffusion coefficients
  diffusion_horiz_coef(param, diff_coef);

  // Compute the diffusion term
  if (IsInterior())
    for (int i=0; i<problem_info.n_diff_horiz; i++)
      {
        VAR ind=problem_info.diff_horiz_ind[i];
        diff[ind][0]=diff_coef[ind][0]*0.5*(Q_dof(0, ind, 0)*basis_func_quad(0, ind)+
                                            Q_dof(1, ind, 0)*basis_func_quad(1, ind));
        diff[ind][1]=diff_coef[ind][1]*0.5*(Q_dof(0, ind, 1)*basis_func_quad(0, ind)+
                                            Q_dof(1, ind, 1)*basis_func_quad(1, ind));
      }
  else
    BC_Neumann_quad(t, diff_coef, diff);

  // Add contribution to the right hand side
  for (int k=0; k<2; k++)
    if (IsElement(k))
      for (int i=0; i<problem_info.n_diff_horiz; i++)
      {
        VAR ind=problem_info.diff_horiz_ind[i];
        L_Var_dof(k, ind).
          Add(-coef_scale(k)*weight_quad()*(diff[ind][0]*Normal(0)+diff[ind][1]*Normal(1)), 
              basis_func_quad(k, ind));
      }
}


//--------------------------------------------------------------------------------------
//                    Vertical velocity computation functions
//--------------------------------------------------------------------------------------

void PRISM_SW::ElementIntegrationVertVelocity()
{
  Var_dof(W).Add(weight_quad()*(Var_dof(U)*basis_func_quad(U)), basis_grad_quad(W, 0));
  Var_dof(W).Add(weight_quad()*(Var_dof(V)*basis_func_quad(V)), basis_grad_quad(W, 1));
}

void PRISM_SW::ElementIntegrationVertVelocityMatrix()
{
  int n_dof=Var_dof(W).GetDim();

  for (int i=0, ind=0; i<n_dof; i++)
    for (int j=0; j<n_dof; j++, ind++)
      w_matrix()[ind]-=weight_quad()*basis_func_quad(W)[j]*basis_grad_quad(W, 2)[i];
}

void FACE_HORIZ_SW::FaceHorizIntegrationVertVelocityTop()
{
  long_real f_hat=(Var_dof(1, U)*basis_func_quad(1, U))*Normal(0)+
    (Var_dof(1, V)*basis_func_quad(1, V))*Normal(1);

  Var_dof(1, W).Add(-coef_scale(1)*weight_quad()*f_hat, basis_func_quad(1, W));
}

void FACE_HORIZ_SW::FaceHorizIntegrationVertVelocityTopMatrix()
{
  int n_dof=Var_dof(1, W).GetDim();
  long_real w_coef=coef_scale(1)*weight_quad()*Normal(2);

  for (int i=0, ind=0; i<n_dof; i++)
    for (int j=0; j<n_dof; j++, ind++)
      w_matrix()[ind]+=w_coef*basis_func_quad(1, W)[i]*basis_func_quad(1, W)[j];
}



//--------------------------------------------------------------------------------------
//                    Baroclinic computation functions
//--------------------------------------------------------------------------------------

#ifdef BAROCLINIC_KEY
void PRISM_SW::ProjectionDensity()
{
  real var[d_var];

  var[XI]=Var_2D_quad(XI);
  var[T]=Var_dof(T)*basis_func_quad(T);
  var[S]=Var_dof(S)*basis_func_quad(S);
  var[RHO]=density(coord_quad(), var);
  Var_dof(RHO).Add(weight_quad()*var[RHO], basis_func_quad(RHO));

  // one- and two-equation turbulence models
#if TURBULENCE_KEY>1
  // Compute the z-component of the density gradient
  RHO_Z_dof().Add(-weight_quad()*var[RHO], basis_grad_quad(RHO_Z, 2));
#endif
}

void FACE_HORIZ_SW::ProjectionDensityFaceHoriz()
{
  density_forcing_quad() = IsSurface() ? 0. : 
    Var_dof(0, RHO)*basis_func_integrated_quad()+Var_dof(0, RHO_FRS)*basis_func_quad_2D(XI);
  if (!IsBottom())
    Var_dof(1, RHO_FRS).Add(container_2d->weight_quad()*density_forcing_quad(), basis_func_quad_2D(XI));

  // one- and two-equation turbulence models
#if TURBULENCE_KEY>1
  real rho = IsInterior() ? 
    0.5*(Var_dof(0, RHO)*basis_func_quad(0, RHO)+Var_dof(1, RHO)*basis_func_quad(1, RHO)) 
    : IsSurface() ? Var_dof(1, RHO)*basis_func_quad(1, RHO) : Var_dof(0, RHO)*basis_func_quad(0, RHO);
  // Compute the z-component of the density gradient
  for (int k=0; k<2; k++)
    if (IsElement(k))
      RHO_Z_dof(k).Add(coef_scale(k)*weight_quad()*rho*Normal(2), basis_func_quad(k, RHO_Z));
#endif
}
#endif



//--------------------------------------------------------------------------------------
//                    Turbulence model functions
//--------------------------------------------------------------------------------------

#if TURBULENCE_KEY>1  // one and two equation turbulence models
//-----------------------------------------
void PRISM_SW::LimitTurbulence()
//-----------------------------------------
{
#if TURBULENCE_KEY==3  // two equation turbulence models
  real k_min=gls_data.k_min;
#else
  real k_min=1.e-6*problem_info.unit_conversion_factor*problem_info.unit_conversion_factor;
#endif
  real k=Var_dof(K)*basis_func_quad(K), d_k = k<k_min ? k_min-k : 0.;
  L_Var_dof(K).Add(weight_quad()*d_k, basis_func_quad(K));
#if TURBULENCE_KEY==3  // two equation turbulence models
  real psi=Var_dof(E)*basis_func_quad(E), d_psi = psi<gls_data.psi_min ? gls_data.psi_min-psi : 0.;
  L_Var_dof(E).Add(weight_quad()*d_psi, basis_func_quad(E));
#endif
}
#endif
