//----------------------------------------------------------------------------------------
// runge_kutta.H: Declaration for the time stepping object
//----------------------------------------------------------------------------------------

#ifndef RUNGE_KUTTA_H
#define RUNGE_KUTTA_H

#include "general.H"

//----------------------------------------------------------------------------------------
// RUNGE_KUTTA class: Time stepping date (Runge-Kutta coefficients)
//----------------------------------------------------------------------------------------
class RUNGE_KUTTA
{
public:
  static const long_real a_table[d_rk][max_stages][max_stages];
  static const long_real b_table[d_rk][max_stages];
  static const long_real d_table[d_rk][max_stages];

  int order;                         // Order of the current Runge-Kutta scheme
  const long_real (*a)[max_stages];  // Pointer to current a-matrix
  const long_real *b;                // Pointer to current b-array (matrix)
  const long_real *d;                // Pointer to current d-array

  // Default constructor
  // Parameters: order_n - order of the time stepping scheme
  RUNGE_KUTTA(int order_n) : order(order_n)
  {
    assert_debug(order>=0 && order<=d_rk);
    a=a_table[order-1];
    b=b_table[order-1];
    d=d_table[order-1];
  }

  // Default destructor
  ~RUNGE_KUTTA(){};

  // Time stepping function
  // Parameters: t_0 - time; dt - time step
  void time_stepping(long_real t_0, long_real dt);
};

// Right-hand side function for the time-stepping routine
// Parameters: t - time
void Func_advection(real t);
// Computation of the auxilliary variables in the mixed formulation
// Parameters: t - time
void Func_diffusion(real t);
// Computation of the vertical velocity component
// Parameters: t - time
void ComputeVerticalVelocity(real t);
// Computation of the baroclinic forcing term
// Parameters: t - time
void ComputeDensityForcing(real t);
// Limiting of minimum values of turbulence variables
// Parameters: t - time
void LimitTurbulenceVariables(real t);

#endif
