//----------------------------------------------------------------------------------------
// Analytical solution for test problems
//----------------------------------------------------------------------------------------

#include <cmath>
#include "pde.H"
#include "test.H"

extern PROBLEM_DESCRIPTION problem_info;
#if TURBULENCE_KEY==3  // two equation turbulence models
extern GLS_Data gls_data;                  // Data for two equation turbulence models
#endif

const real amp=0.01;

// True solution for the test problem
void solution_analytic(real t, const real coord[d_space], real func[d_var])
{
  func[U]=amp*coord[1];
  func[V]=0.;
  func[W]=0.;
#ifdef BAROCLINIC_KEY
  func[T]=20.;
  func[S]=35.;
//   func[RO]=density(problem_info, coord, func);
#endif
}

// Gradient of the true solution
void solution_analytic_grad(real t, const real coord[d_space], real grad[d_var][d_space])
{
  grad[U][0]=0.;
  grad[U][1]=amp;
  grad[U][2]=0.;
  grad[V][0]=0.;
  grad[V][1]=0.;
  grad[V][2]=0.;
  grad[W][0]=0.;
  grad[W][1]=0.;
  grad[W][2]=0.;
#ifdef BAROCLINIC_KEY
  grad[T][0]=0.;
  grad[T][1]=0.;
  grad[T][2]=0.;
  grad[S][0]=0.;
  grad[S][1]=0.;
  grad[S][2]=0.;
#endif
}

// Laplacian of the true solution
void solution_analytic_laplacian(real t, const real coord[d_space], real lapl[d_var][d_space])
{
  lapl[U][0]=0.;
  lapl[U][1]=0.;
  lapl[U][2]=0.;
  lapl[V][0]=0.;
  lapl[V][1]=0.;
  lapl[V][2]=0.;
#ifdef BAROCLINIC_KEY
  lapl[T][0]=0.;
  lapl[T][1]=0.;
  lapl[T][2]=0.;
  lapl[S][0]=0.;
  lapl[S][1]=0.;
  lapl[S][2]=0.;
#endif
}

// Time derivative of the true solution for the test problem
void solution_analytic_dt(real t, const real coord[d_space], real t_deriv[d_var])
{
  t_deriv[U]=0.;
  t_deriv[V]=0.;
#ifdef BAROCLINIC_KEY
  t_deriv[T]=0.;
  t_deriv[S]=0.;
#endif
}

// Analytic solution for the elevation
void solution_analytic_elevation(real t, const real coord[d_space_2d], real result[])
{
  result[0]=amp*coord[0];
}

// Analytic solution for the elevation gradient
void solution_analytic_elevation_grad(real t, const real coord[d_space_2d], real elev_grad[d_space_2d])
{
  elev_grad[0]=amp;
  elev_grad[1]=0.;
}

// Time derivative of the true solution for the elevation
void solution_analytic_elevation_dt(real t, const real coord[d_space], real result[])
{
  result[0]=0.;
}

#ifdef BAROCLINIC_KEY
long_real density(const real coord[d_space], const real var[d_var])
{
  return 0.1*coord[0];
}

// Forcing function accounting for density terms
void density_forcing(real t, real xi, const real xi_grad[d_space_2d], const real coord[d_space], real result[])
{
  real coord_top[d_space]={coord[0], coord[1], xi}, var[d_var], rho=density(coord_top, var);

  result[0]=problem_info.g*(0.1*(xi-coord[2])+rho*xi_grad[0]);
  result[1]=problem_info.g*rho*xi_grad[1];
}
#endif

// Vertical eddy viscosity/diffusion computation
void diffusion_vert_coef(real t, const real coord[d_space], const real var[d_var], 
                         const real grad[d_var][d_space], real diff_coef[d_var][d_space], 
                         real diff_coef_grad[d_var][d_space], real rhs[d_var])
{
  const real kappa=0.4, u_av=1., H=10.;//, u_star=0.0061;
  real z_c=problem_info.bottom_rough_coef*std::exp(kappa/std::sqrt(problem_info.bottom_fric_coef));
  real param[6]={0.};

  switch(problem_info.diff_vert_switch)
    {
    case 0: // No verical eddy viscosity
      break;
    case 1: // Constant vertical eddy viscosity coeffficient
      diffusion_vert_coef_const(param, diff_coef);
      break;
#if TURBULENCE_KEY>1  // one equation turbulence models
    case 4: // Delft3D
#ifdef BAROCLINIC_KEY
      param[0]=problem_info.g/problem_info.Ro_0*density_grad(t, coord);
#else
      param[0]=0.;
#endif
      param[1]=var[K];
      param[2]=-coord[2];
      param[3]=10.+coord[2];
      param[4]=grad[U][2];
      param[5]=grad[V][2];
      diffusion_vert_coef_k(param, diff_coef);
//       diff_coef_grad[U][2]=diff_coef_grad[V][2]=diff_coef_grad[K][2]=0.00244*(-10.-2.*coord[2]);
      diff_coef_grad[U][2]=diff_coef_grad[V][2]=diff_coef_grad[K][2]=0.00244*(-10.-2.*coord[2]-z_c);

      // Add the right hand side contribution for the k model
      rhs_k(param, rhs);
      break;
#if TURBULENCE_KEY==3
  case 6: // Burchard et al. k-eps turbulence model
#ifdef BAROCLINIC_KEY
    param[0]=problem_info.g/problem_info.Ro_0*density_grad(t, coord);
#else
    param[0]=0.;
#endif
    param[1]=var[K];
    param[2]=var[E];
    param[3]=1.;
    param[4]=grad[U][2];
    param[5]=grad[V][2];
    diffusion_vert_coef_gls(param, diff_coef);
    diff_coef_grad[U][2]=diff_coef_grad[V][2]=
      gls_data.C*var[K]*(2.*grad[K][2]*var[E]-var[K]*grad[E][2])/var[E]/var[E];
    diff_coef_grad[K][2]=diff_coef_grad[U][2]/gls_data.sigma_k;
    diff_coef_grad[E][2]=diff_coef_grad[U][2]/gls_data.sigma_psi;
//     rhs_gls(param, rhs);
    real u_term=grad[U][2]*grad[U][2]+grad[V][2]*grad[V][2];
    rhs[K]=problem_info.ramp_value*(diff_coef[U][2]*u_term-var[E]);
    rhs[E]=problem_info.ramp_value*var[E]/var[K]*(gls_data.c1*diff_coef[U][2]*u_term-gls_data.c2*var[E]);
    break;
#endif
#endif
    default: std::cerr << "File test_*: Unknown vertical eddy viscosity model choice parameter!\n";
      std::exit(1);
    }
}
